@@ -20,24 +20,25 @@ interact with UNIX and Windows password database.
 
 import errno
 import os
-import sys
 import warnings
 
 from ._compat import PY3
-from ._compat import unicode
 from ._compat import getcwdu
+from ._compat import unicode
 
 
-__all__ = ['DummyAuthorizer',
-           # 'BaseUnixAuthorizer', 'UnixAuthorizer',
-           # 'BaseWindowsAuthorizer', 'WindowsAuthorizer',
-           ]
+__all__ = [
+    'DummyAuthorizer',
+    # 'BaseUnixAuthorizer', 'UnixAuthorizer',
+    # 'BaseWindowsAuthorizer', 'WindowsAuthorizer',
+]
 
 
 # ===================================================================
 # --- exceptions
 # ===================================================================
 
+
 class AuthorizerError(Exception):
     """Base class for authorizer exceptions."""
 
@@ -50,7 +51,8 @@ class AuthenticationFailed(Exception):
 # --- base class
 # ===================================================================
 
-class DummyAuthorizer(object):
+
+class DummyAuthorizer:
     """Basic "dummy" authorizer class, suitable for subclassing to
     create your own custom authorizers.
 
@@ -67,13 +69,20 @@ class DummyAuthorizer(object):
     """
 
     read_perms = "elr"
-    write_perms = "adfmwM"
+    write_perms = "adfmwMT"
 
     def __init__(self):
         self.user_table = {}
 
-    def add_user(self, username, password, homedir, perm='elr',
-                 msg_login="Login successful.", msg_quit="Goodbye."):
+    def add_user(
+        self,
+        username,
+        password,
+        homedir,
+        perm='elr',
+        msg_login="Login successful.",
+        msg_quit="Goodbye.",
+    ):
         """Add a user to the virtual users table.
 
         AuthorizerError exceptions raised on error conditions such as
@@ -94,6 +103,7 @@ class DummyAuthorizer(object):
          - "m" = create directory (MKD command)
          - "w" = store a file to the server (STOR, STOU commands)
          - "M" = change file mode (SITE CHMOD command)
+         - "T" = update file last modified time (MFMT command)
 
         Optional msg_login and msg_quit arguments can be specified to
         provide customized response strings when user log-in and quit.
@@ -106,13 +116,14 @@ class DummyAuthorizer(object):
             raise ValueError('no such directory: %r' % homedir)
         homedir = os.path.realpath(homedir)
         self._check_permissions(username, perm)
-        dic = {'pwd': str(password),
-               'home': homedir,
-               'perm': perm,
-               'operms': {},
-               'msg_login': str(msg_login),
-               'msg_quit': str(msg_quit)
-               }
+        dic = {
+            'pwd': str(password),
+            'home': homedir,
+            'perm': perm,
+            'operms': {},
+            'msg_login': str(msg_login),
+            'msg_quit': str(msg_quit),
+        }
         self.user_table[username] = dic
 
     def add_anonymous(self, homedir, **kwargs):
@@ -199,19 +210,22 @@ class DummyAuthorizer(object):
         pathname of a file or a directory).
 
         Expected perm argument is one of the following letters:
-        "elradfmwM".
+        "elradfmwMT".
         """
         if path is None:
             return perm in self.user_table[username]['perm']
 
         path = os.path.normcase(path)
-        for dir in self.user_table[username]['operms'].keys():
+        for dir in self.user_table[username]['operms']:
             operm, recursive = self.user_table[username]['operms'][dir]
             if self._issubpath(path, dir):
                 if recursive:
                     return perm in operm
-                if (path == dir or os.path.dirname(path) == dir and not
-                        os.path.isdir(path)):
+                if (
+                    path == dir
+                    or os.path.dirname(path) == dir
+                    and not os.path.isdir(path)
+                ):
                     return perm in operm
 
         return perm in self.user_table[username]['perm']
@@ -236,18 +250,23 @@ class DummyAuthorizer(object):
         for p in perm:
             if p not in self.read_perms + self.write_perms:
                 raise ValueError('no such permission %r' % p)
-            if (username == 'anonymous' and
-                    p in self.write_perms and not
-                    warned):
-                warnings.warn("write permissions assigned to anonymous user.",
-                              RuntimeWarning)
+            if (
+                username == 'anonymous'
+                and p in self.write_perms
+                and not warned
+            ):
+                warnings.warn(
+                    "write permissions assigned to anonymous user.",
+                    RuntimeWarning,
+                    stacklevel=2,
+                )
                 warned = 1
 
     def _issubpath(self, a, b):
         """Return True if a is a sub-path of b or if the paths are equal."""
         p1 = a.rstrip(os.sep).split(os.sep)
         p2 = b.rstrip(os.sep).split(os.sep)
-        return p1[:len(p2)] == p2
+        return p1[: len(p2)] == p2
 
 
 def replace_anonymous(callable):
@@ -260,6 +279,7 @@ def replace_anonymous(callable):
         if username == 'anonymous':
             username = self.anonymous_user or username
         return callable(self, username, *args, **kwargs)
+
     return wrapper
 
 
@@ -267,7 +287,8 @@ def replace_anonymous(callable):
 # --- platform specific authorizers
 # ===================================================================
 
-class _Base(object):
+
+class _Base:
     """Methods common to both Unix and Windows authorizers.
     Not supposed to be used directly.
     """
@@ -281,11 +302,13 @@ class _Base(object):
     def __init__(self):
         """Check for errors in the constructor."""
         if self.rejected_users and self.allowed_users:
-            raise AuthorizerError("rejected_users and allowed_users options "
-                                  "are mutually exclusive")
+            raise AuthorizerError(
+                "rejected_users and allowed_users options "
+                "are mutually exclusive"
+            )
 
         users = self._get_system_users()
-        for user in (self.allowed_users or self.rejected_users):
+        for user in self.allowed_users or self.rejected_users:
             if user == 'anonymous':
                 raise AuthorizerError('invalid username "anonymous"')
             if user not in users:
@@ -296,18 +319,32 @@ class _Base(object):
                 raise AuthorizerError('no such user %s' % self.anonymous_user)
             home = self.get_home_dir(self.anonymous_user)
             if not os.path.isdir(home):
-                raise AuthorizerError('no valid home set for user %s'
-                                      % self.anonymous_user)
+                raise AuthorizerError(
+                    'no valid home set for user %s' % self.anonymous_user
+                )
 
-    def override_user(self, username, password=None, homedir=None, perm=None,
-                      msg_login=None, msg_quit=None):
+    def override_user(
+        self,
+        username,
+        password=None,
+        homedir=None,
+        perm=None,
+        msg_login=None,
+        msg_quit=None,
+    ):
         """Overrides the options specified in the class constructor
         for a specific user.
         """
-        if (not password and not homedir and not perm and not msg_login and not
-                msg_quit):
+        if (
+            not password
+            and not homedir
+            and not perm
+            and not msg_login
+            and not msg_quit
+        ):
             raise AuthorizerError(
-                "at least one keyword argument must be specified")
+                "at least one keyword argument must be specified"
+            )
         if self.allowed_users and username not in self.allowed_users:
             raise AuthorizerError('%s is not an allowed user' % username)
         if self.rejected_users and username in self.rejected_users:
@@ -322,12 +359,14 @@ class _Base(object):
         if username in self._dummy_authorizer.user_table:
             # re-set parameters
             del self._dummy_authorizer.user_table[username]
-        self._dummy_authorizer.add_user(username,
-                                        password or "",
-                                        homedir or getcwdu(),
-                                        perm or "",
-                                        msg_login or "",
-                                        msg_quit or "")
+        self._dummy_authorizer.add_user(
+            username,
+            password or "",
+            homedir or getcwdu(),
+            perm or "",
+            msg_login or "",
+            msg_quit or "",
+        )
         if homedir is None:
             self._dummy_authorizer.user_table[username]['home'] = ""
 
@@ -358,9 +397,7 @@ class _Base(object):
         """
         if self.allowed_users and username not in self.allowed_users:
             return True
-        if self.rejected_users and username in self.rejected_users:
-            return True
-        return False
+        return bool(self.rejected_users and username in self.rejected_users)
 
 
 # ===================================================================
@@ -380,7 +417,7 @@ else:
     PROCESS_UID = os.getuid()
     PROCESS_GID = os.getgid()
 
-    class BaseUnixAuthorizer(object):
+    class BaseUnixAuthorizer:
         """An authorizer compatible with Unix user account and password
         database.
         This class should not be used directly unless for subclassing.
@@ -394,7 +431,7 @@ else:
 
             if self.anonymous_user is not None:
                 try:
-                    pwd.getpwnam(self.anonymous_user).pw_dir
+                    pwd.getpwnam(self.anonymous_user).pw_dir  # noqa
                 except KeyError:
                     raise AuthorizerError('no such user %s' % anonymous_user)
 
@@ -469,7 +506,7 @@ else:
             return "Goodbye."
 
         def get_perms(self, username):
-            return "elradfmw"
+            return "elradfmwMT"
 
         def has_perm(self, username, perm, path=None):
             return perm in self.get_perms(username)
@@ -497,47 +534,50 @@ else:
 
         # --- public API
 
-        def __init__(self, global_perm="elradfmw",
-                     allowed_users=None,
-                     rejected_users=None,
-                     require_valid_shell=True,
-                     anonymous_user=None,
-                     msg_login="Login successful.",
-                     msg_quit="Goodbye."):
+        def __init__(
+            self,
+            global_perm="elradfmwMT",
+            allowed_users=None,
+            rejected_users=None,
+            require_valid_shell=True,
+            anonymous_user=None,
+            msg_login="Login successful.",
+            msg_quit="Goodbye.",
+        ):
             """Parameters:
 
-             - (string) global_perm:
-                a series of letters referencing the users permissions;
-                defaults to "elradfmw" which means full read and write
-                access for everybody (except anonymous).
+            - (string) global_perm:
+               a series of letters referencing the users permissions;
+               defaults to "elradfmwMT" which means full read and write
+               access for everybody (except anonymous).
 
-             - (list) allowed_users:
-                a list of users which are accepted for authenticating
-                against the FTP server; defaults to [] (no restrictions).
+            - (list) allowed_users:
+               a list of users which are accepted for authenticating
+               against the FTP server; defaults to [] (no restrictions).
 
-             - (list) rejected_users:
-                a list of users which are not accepted for authenticating
-                against the FTP server; defaults to [] (no restrictions).
+            - (list) rejected_users:
+               a list of users which are not accepted for authenticating
+               against the FTP server; defaults to [] (no restrictions).
 
-             - (bool) require_valid_shell:
-                Deny access for those users which do not have a valid shell
-                binary listed in /etc/shells.
-                If /etc/shells cannot be found this is a no-op.
-                Anonymous user is not subject to this option, and is free
-                to not have a valid shell defined.
-                Defaults to True (a valid shell is required for login).
+            - (bool) require_valid_shell:
+               Deny access for those users which do not have a valid shell
+               binary listed in /etc/shells.
+               If /etc/shells cannot be found this is a no-op.
+               Anonymous user is not subject to this option, and is free
+               to not have a valid shell defined.
+               Defaults to True (a valid shell is required for login).
 
-             - (string) anonymous_user:
-                specify it if you intend to provide anonymous access.
-                The value expected is a string representing the system user
-                to use for managing anonymous sessions;  defaults to None
-                (anonymous access disabled).
+            - (string) anonymous_user:
+               specify it if you intend to provide anonymous access.
+               The value expected is a string representing the system user
+               to use for managing anonymous sessions;  defaults to None
+               (anonymous access disabled).
 
-             - (string) msg_login:
-                the string sent when client logs in.
+            - (string) msg_login:
+               the string sent when client logs in.
 
-             - (string) msg_quit:
-                the string sent when client quits.
+            - (string) msg_quit:
+               the string sent when client quits.
             """
             BaseUnixAuthorizer.__init__(self, anonymous_user)
             if allowed_users is None:
@@ -558,19 +598,28 @@ else:
             if require_valid_shell:
                 for username in self.allowed_users:
                     if not self._has_valid_shell(username):
-                        raise AuthorizerError("user %s has not a valid shell"
-                                              % username)
+                        raise AuthorizerError(
+                            "user %s has not a valid shell" % username
+                        )
 
-        def override_user(self, username, password=None, homedir=None,
-                          perm=None, msg_login=None, msg_quit=None):
+        def override_user(
+            self,
+            username,
+            password=None,
+            homedir=None,
+            perm=None,
+            msg_login=None,
+            msg_quit=None,
+        ):
             """Overrides the options specified in the class constructor
             for a specific user.
             """
             if self.require_valid_shell and username != 'anonymous':
                 if not self._has_valid_shell(username):
                     raise AuthorizerError(self.msg_invalid_shell % username)
-            _Base.override_user(self, username, password, homedir, perm,
-                                msg_login, msg_quit)
+            _Base.override_user(
+                self, username, password, homedir, perm, msg_login, msg_quit
+            )
 
         # --- overridden / private API
 
@@ -586,12 +635,14 @@ else:
                 if overridden_password != password:
                     raise AuthenticationFailed(self.msg_wrong_password)
             else:
-                BaseUnixAuthorizer.validate_authentication(self, username,
-                                                           password, handler)
+                BaseUnixAuthorizer.validate_authentication(
+                    self, username, password, handler
+                )
             if self.require_valid_shell and username != 'anonymous':
                 if not self._has_valid_shell(username):
                     raise AuthenticationFailed(
-                        self.msg_invalid_shell % username)
+                        self.msg_invalid_shell % username
+                    )
 
         @replace_anonymous
         def has_user(self, username):
@@ -612,7 +663,7 @@ else:
             in /etc/shells. If /etc/shells can't be found return True.
             """
             try:
-                file = open('/etc/shells', 'r')
+                file = open('/etc/shells')
             except IOError as err:
                 if err.errno == errno.ENOENT:
                     return True
@@ -645,15 +696,15 @@ try:
     import win32security
 except ImportError:
     pass
-else:
-    if sys.version_info < (3, 0):
-        import _winreg as winreg
-    else:
+else:  # pragma: no cover
+    if PY3:
         import winreg
+    else:
+        import _winreg as winreg
 
     __all__.extend(['BaseWindowsAuthorizer', 'WindowsAuthorizer'])
 
-    class BaseWindowsAuthorizer(object):
+    class BaseWindowsAuthorizer:
         """An authorizer compatible with Windows user account and
         password database.
         This class should not be used directly unless for subclassing.
@@ -665,8 +716,9 @@ else:
             self.anonymous_user = anonymous_user
             self.anonymous_password = anonymous_password
             if self.anonymous_user is not None:
-                self.impersonate_user(self.anonymous_user,
-                                      self.anonymous_password)
+                self.impersonate_user(
+                    self.anonymous_user, self.anonymous_password
+                )
                 self.terminate_impersonation(None)
 
         def validate_authentication(self, username, password, handler):
@@ -675,9 +727,13 @@ else:
                     raise AuthenticationFailed(self.msg_anon_not_allowed)
                 return
             try:
-                win32security.LogonUser(username, None, password,
-                                        win32con.LOGON32_LOGON_INTERACTIVE,
-                                        win32con.LOGON32_PROVIDER_DEFAULT)
+                win32security.LogonUser(
+                    username,
+                    None,
+                    password,
+                    win32con.LOGON32_LOGON_INTERACTIVE,
+                    win32con.LOGON32_PROVIDER_DEFAULT,
+                )
             except pywintypes.error:
                 raise AuthenticationFailed(self.msg_wrong_password)
 
@@ -685,9 +741,12 @@ else:
         def impersonate_user(self, username, password):
             """Impersonate the security context of another user."""
             handler = win32security.LogonUser(
-                username, None, password,
+                username,
+                None,
+                password,
                 win32con.LOGON32_LOGON_INTERACTIVE,
-                win32con.LOGON32_PROVIDER_DEFAULT)
+                win32con.LOGON32_PROVIDER_DEFAULT,
+            )
             win32security.ImpersonateLoggedOnUser(handler)
             handler.Close()
 
@@ -706,16 +765,18 @@ else:
             """
             try:
                 sid = win32security.ConvertSidToStringSid(
-                    win32security.LookupAccountName(None, username)[0])
+                    win32security.LookupAccountName(None, username)[0]
+                )
             except pywintypes.error as err:
                 raise AuthorizerError(err)
-            path = r"SOFTWARE\Microsoft\Windows NT" \
-                   r"\CurrentVersion\ProfileList" + "\\" + sid
+            path = r"SOFTWARE\Microsoft\Windows NT"
+            path += r"\CurrentVersion\ProfileList" + "\\" + sid
             try:
                 key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, path)
             except WindowsError:
                 raise AuthorizerError(
-                    "No profile directory defined for user %s" % username)
+                    "No profile directory defined for user %s" % username
+                )
             value = winreg.QueryValueEx(key, "ProfileImagePath")[0]
             home = win32api.ExpandEnvironmentStrings(value)
             if not PY3 and not isinstance(home, unicode):
@@ -727,8 +788,9 @@ else:
             """Return all users defined on the Windows system."""
             # XXX - Does Windows allow usernames with chars outside of
             # ASCII set? In that case we need to convert this to unicode.
-            return [entry['name'] for entry in
-                    win32net.NetUserEnum(None, 0)[0]]
+            return [
+                entry['name'] for entry in win32net.NetUserEnum(None, 0)[0]
+            ]
 
         def get_msg_login(self, username):
             return "Login successful."
@@ -737,7 +799,7 @@ else:
             return "Goodbye."
 
         def get_perms(self, username):
-            return "elradfmw"
+            return "elradfmwMT"
 
         def has_perm(self, username, perm, path=None):
             return perm in self.get_perms(username)
@@ -762,47 +824,49 @@ else:
 
         # --- public API
 
-        def __init__(self,
-                     global_perm="elradfmw",
-                     allowed_users=None,
-                     rejected_users=None,
-                     anonymous_user=None,
-                     anonymous_password=None,
-                     msg_login="Login successful.",
-                     msg_quit="Goodbye."):
+        def __init__(
+            self,
+            global_perm="elradfmwMT",
+            allowed_users=None,
+            rejected_users=None,
+            anonymous_user=None,
+            anonymous_password=None,
+            msg_login="Login successful.",
+            msg_quit="Goodbye.",
+        ):
             """Parameters:
 
-             - (string) global_perm:
-                a series of letters referencing the users permissions;
-                defaults to "elradfmw" which means full read and write
-                access for everybody (except anonymous).
+            - (string) global_perm:
+               a series of letters referencing the users permissions;
+               defaults to "elradfmwMT" which means full read and write
+               access for everybody (except anonymous).
 
-             - (list) allowed_users:
-                a list of users which are accepted for authenticating
-                against the FTP server; defaults to [] (no restrictions).
+            - (list) allowed_users:
+               a list of users which are accepted for authenticating
+               against the FTP server; defaults to [] (no restrictions).
 
-             - (list) rejected_users:
-                a list of users which are not accepted for authenticating
-                against the FTP server; defaults to [] (no restrictions).
+            - (list) rejected_users:
+               a list of users which are not accepted for authenticating
+               against the FTP server; defaults to [] (no restrictions).
 
-             - (string) anonymous_user:
-                specify it if you intend to provide anonymous access.
-                The value expected is a string representing the system user
-                to use for managing anonymous sessions.
-                As for IIS, it is recommended to use Guest account.
-                The common practice is to first enable the Guest user, which
-                is disabled by default and then assign an empty password.
-                Defaults to None (anonymous access disabled).
+            - (string) anonymous_user:
+               specify it if you intend to provide anonymous access.
+               The value expected is a string representing the system user
+               to use for managing anonymous sessions.
+               As for IIS, it is recommended to use Guest account.
+               The common practice is to first enable the Guest user, which
+               is disabled by default and then assign an empty password.
+               Defaults to None (anonymous access disabled).
 
-             - (string) anonymous_password:
-                the password of the user who has been chosen to manage the
-                anonymous sessions.  Defaults to None (empty password).
+            - (string) anonymous_password:
+               the password of the user who has been chosen to manage the
+               anonymous sessions.  Defaults to None (empty password).
 
-             - (string) msg_login:
-                the string sent when client logs in.
+            - (string) msg_login:
+               the string sent when client logs in.
 
-             - (string) msg_quit:
-                the string sent when client quits.
+            - (string) msg_quit:
+               the string sent when client quits.
             """
             if allowed_users is None:
                 allowed_users = []
@@ -820,17 +884,26 @@ else:
             _Base.__init__(self)
             # actually try to impersonate the user
             if self.anonymous_user is not None:
-                self.impersonate_user(self.anonymous_user,
-                                      self.anonymous_password)
+                self.impersonate_user(
+                    self.anonymous_user, self.anonymous_password
+                )
                 self.terminate_impersonation(None)
 
-        def override_user(self, username, password=None, homedir=None,
-                          perm=None, msg_login=None, msg_quit=None):
+        def override_user(
+            self,
+            username,
+            password=None,
+            homedir=None,
+            perm=None,
+            msg_login=None,
+            msg_quit=None,
+        ):
             """Overrides the options specified in the class constructor
             for a specific user.
             """
-            _Base.override_user(self, username, password, homedir, perm,
-                                msg_login, msg_quit)
+            _Base.override_user(
+                self, username, password, homedir, perm, msg_login, msg_quit
+            )
 
         # --- overridden / private API
 
@@ -853,7 +926,8 @@ else:
                     raise AuthenticationFailed(self.msg_wrong_password)
             else:
                 BaseWindowsAuthorizer.validate_authentication(
-                    self, username, password, handler)
+                    self, username, password, handler
+                )
 
         def impersonate_user(self, username, password):
             """Impersonate the security context of another user."""
